(in-package :iolib.examples)

(defvar *ex5-server-client-threads* nil)
(defvar *ex5-tls-client* nil)

(defun run-ex5-server-helper (port)
  (with-open-socket
      (server :connect :passive
              :address-family :internet
              :type :stream
              :ipv6 nil
              :external-format '(:utf-8 :eol-style :crlf))

    (format t "Created socket: ~A[fd=~A]~%" server (socket-os-fd server))

    ;; Bind the socket to all interfaces with specified port.
    (bind-address server +ipv4-unspecified+
                  :port port
                  :reuse-addr t)
    (format t "Bound socket: ~A~%" server)

    ;; start listening on the server socket
    (listen-on server :backlog 5)
    (format t "Listening on socket bound to: ~A:~A~%"
            (local-name server)
            (local-port server))

    ;; keep accepting connections forever, but if this exits for whatever
    ;; reason ensure to destroy any remaining running threads.
    (unwind-protect
         (do ()
             (nil nil)
           ;; *ex5-tls-client* is the dynamic variable which will be bound
           ;; to whatever I need the thread to know when it wakes up
           (declare (special *ex5-tls-client*))
           ;; set up a thread to handle the client.
           (format t "Waiting to accept a connection...~%")
           (let* ((client (accept-connection server :wait t))
                  (*default-special-bindings*
                   `((*ex5-tls-client* . ,(funcall (lambda () client)))
                     ,@*default-special-bindings*)))

             ;; keep track of any active threads we currently know about...
             (setf *ex5-server-client-threads*
                   (remove-if-not #'thread-alive-p
                                  *ex5-server-client-threads*))
             (push
              ;; ...and handle the connection!
              (make-thread #'process-ex5-client-thread)
              *ex5-server-client-threads*)))

      ;; clean up all of the client threads when done.
      (format t "Destroying any active threads....~%")
      (mapc #'(lambda (thr)
                ;; XXX What do I do about the race between the check and
                ;; the destroy?
                (when (thread-alive-p thr)
                  (format t "Destroying thread: ~A~%" thr)
                  (destroy-thread thr)))
            *ex5-server-client-threads*)
      (setf *ex5-server-client-threads* nil))))

;;; The thread which handles the client connection.
(defun process-ex5-client-thread ()
  ;; declared ignorable because it is set out of the context of this function.
  (declare (ignorable *ex5-tls-client*))
  (let ((client *ex5-tls-client*)
        (done nil))
    (multiple-value-bind (remote rport)
        (remote-name client)
      (format t "A thread is handling the connection from ~A:~A!~%"
	      remote rport)

      ;; no matter how we get out of the client processing loop, we always
      ;; close the connection.
      (unwind-protect
           (handler-case
               ;;  perform the actual echoing algorithm
               (str-echo client remote rport)

             (end-of-file ()
               (format t "Client ~A:~A closed connection for a read.~%"
                       remote rport)
               t)

             (hangup ()
               (format t "Client ~A:~A closed connection for a write.~%"
                       remote rport)
               t))

        ;; firing action of the unwind-protect
        (format t "Closing connection to ~A:~A~%!" remote rport)
        (close client)
        t))))

(defun str-echo (client remote rport)
  (do ()
      ;; here we let signaled conditions on the boundary conditions of the
      ;; client (meaning it closes its connection to us on either a read or
      ;; a write) bail us out of the infinite loop
      (nil nil)
    ;; Since we're using an internet TCP stream,
    ;; we can use format with it. However, we
    ;; should be sure to finish-output in order that
    ;; all the data is sent.
    (let ((line (read-line client)))
      (format t "Read line from ~A:~A: ~A~%" remote rport line)
      (format client "~A~%" line)
      (finish-output client)
      (format t "Wrote line to ~A:~A: ~A~%" remote rport line)
      t)))

;; This just checks for some error conditions so we can print out a nice
;; message about it.
(defun run-ex5-server (&key (port *port*))
  (handler-case
      (run-ex5-server-helper port)
    ;; handle some common conditions
    (socket-address-in-use-error ()
      (format t "Bind: Address already in use, forget :reuse-addr t?"))))



