(in-package :iolib.examples)

;; This function returns an closure from which I can ask for a
;; function that I can use as a reader on a socket which reads lines,
;; and a function that I can use as a writer on a socket to write
;; lines stored in the buffer onto the socket.  When constructing the
;; buffer, I specify the socket upon which the buffer is responsible
;; and how bytes I'm willing to buffer. Obviously, if a line is very
;; much bigger than the max-bytes I have, I have to read the whole
;; thing due to read-line.

(defun make-ex7-io-buffer (socket who port disconnector &key (max-bytes 4096))
  (declare (special *ex7-event-base*))
  (let ((line-queue (make-queue))
        (bytes-left-to-write 0)
        (read-handler-registered nil)
        (write-handler-registered nil))
    (labels
        ;; If this function notices that there is data to write, it will
        ;; set the io-handler on the socket for the write handler.
        ;; If the function notices it has read >= than the max-bytes
        ;; it will remove itself from the handler *after* ensuring the
        ;; write handler is set up properly.
        ((read-a-line (fd event exception)
           (handler-case
               (let ((line (format nil "~A~%" (read-line socket)))) ; add a \n
                 (format t "Read from ~A:~A: ~A" who port line)
                 (enqueue line line-queue)
                 (incf bytes-left-to-write (length line))
                 (when (> bytes-left-to-write 0)
                   ;; If the write handler isn't registered, then do
                   ;; it now since I have data to write.
                   (unless write-handler-registered
                     (set-io-handler *ex7-event-base*
                                     (socket-os-fd socket)
                                     :write
                                     #'write-a-line)
                     (setf write-handler-registered t)))
                 ;; Now, if there is more data than I should be
                 ;; reading, remove myself from the io handler. When
                 ;; the write handler notices that, after writing some
                 ;; data, more of it can be read, it will reregister
                 ;; the io handler for the read socket.
                 (when (>= bytes-left-to-write max-bytes)
                   (remove-fd-handlers *ex7-event-base*
                                       (socket-os-fd socket)
                                       :read t)
                   (setf read-handler-registered nil)))

             (end-of-file ()
               (format t "Client ~A:~A produced end-of-file on a read.~%"
                       who port)
               (funcall disconnector who port))))

         ;; This function will notice that if it has written enough bytes to
         ;; bring the bytes-left-to-write under max-bytes, it will re-register
         ;; the reader io handler. If there is no data to write, it will,
         ;; after ensuring the read handler is registered, unregister itself
         ;; as to not be called constantly on a write ready socket with no
         ;; data to write.
         (write-a-line (fd event exception)
           (handler-case
               (progn
                 (when (> bytes-left-to-write 0)
                   (let ((line (dequeue line-queue)))
                     (format socket "~A" line) ;; newline is in the string.
                     (finish-output socket)
                     (format t "Wrote to ~A:~A: ~A" who port line)
                     (decf bytes-left-to-write (length line))))

                 ;; If we see we've fallen below the max-bytes mark,
                 ;; re-register the read handler to get more data for
                 ;; us.
                 (when (< bytes-left-to-write max-bytes)
                   (unless read-handler-registered
                     (set-io-handler *ex7-event-base*
                                     (socket-os-fd socket)
                                     :read
                                     #'read-a-line)
                     (setf read-handler-registered t)))

                 ;; If we happen to notice we've written all of our data, then
                 ;; unregister the write handler so we don't get called
                 ;; unnecesarily. This method might mean that sometimes we'll have
                 ;; to make an extra trip through the event-dispatcher to perform
                 ;; the write.
                 (when (zerop bytes-left-to-write)
                   (remove-fd-handlers *ex7-event-base*
                                       (socket-os-fd socket)
                                       :write t)
                   (setf write-handler-registered nil)))

             (hangup ()
               (format t "Client ~A:~A got hangup on write.~%" who port)
               (funcall disconnector who port)))))

      ;; This is the actual function returned from make-ex7-io-buffer
      ;; which allows us access to the read/writer in the scope of the
      ;; closure.  We will ask for the correct functions when setting
      ;; up the io handlers.  NOTE: By simply asking for the handler,
      ;; I've assumed it is to be immediately put into an iolib event
      ;; handler. This is why they are considered registered at this point.
      (lambda (msg)
        (cond
          ((equalp msg :read-a-line)
           (setf read-handler-registered t)
           #'read-a-line)
          ((equalp msg :write-a-line)
           (setf write-handler-registered t)
           #'write-a-line)
          (t
           (error "make-ex7-buffer: Please supply :read-a-line or :write-a-line~%")))))))


;; a small, but efficient, queue implementation.
(defun make-queue ()
  (cons nil nil))

(defun enqueue (obj q)
  (if (null (car q))
      (setf (cdr q) (setf (car q) (list obj)))
      (setf (cdr (cdr q)) (list obj)
            (cdr q) (cdr (cdr q))))

  (car q))

(defun dequeue (q)
  (pop (car q)))
